from setuptools import setup

setup(
    name='aiopipes',
    version='0.3',
    packages=['aiopipes'],
    url='',
    license='',
    author='Orf',
    author_email='tom@tomforb.es',
    description='Asyncio pipe structures',
    install_requires=["asyncio", "decorator"],
)
